@@ -4,6 +4,7 @@
 {% endblock %}
 
 {% block scheduleLoop %}
+    CG_BEFORE_SCHEDULE;
 {% if config.debug %}
     while((cgStaticError==0) && (debugCounter > 0))
 {% else %}
@@ -19,7 +20,9 @@
        {% if config.eventRecorder -%}
        EventRecord2 (Evt_Node, {{nodes[s].codeID}}, 0);
        {% endif -%}
+       CG_BEFORE_NODE_EXECUTION;
        {{nodes[s].cRun()}}
+       CG_AFTER_NODE_EXECUTION;
        {% if config.eventRecorder -%}
        if (cgStaticError<0)
        {
